import unittest
import base64
from typing import Optional, Any
import neo3crypto
from aioresponses import aioresponses
from neo3 import api
from neo3.network.payloads import transaction, block, verification
from neo3.core import types, cryptography
from neo3.wallet import utils

JSON = Any


class TestNeoRpcClient(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.client = api.NeoRpcClient("localhost")
        # CAREFULL THIS PATCHES ALL aiohttp CALLS!
        self.helper = aioresponses()
        self.helper.start()

    async def asyncTearDown(self) -> None:
        self.helper.stop()
        await self.client.close()

    def mock_response(
        self, payload: Optional[JSON] = None, exc: Optional[Exception] = None
    ):
        """
        Either payload or exc should be provided
        """
        if payload is not None and exc is not None:
            raise ValueError("Arguments are mutual exclusive")

        if payload is not None:
            json = {"jsonrpc": "2.0", "id": 1, "result": payload}
            self.helper.post("localhost", payload=json)
        else:
            self.helper.post("localhost", exception=exc)

    async def test_calculate_network_fee(self):
        self.mock_response({"networkfee": "123"})
        response = await self.client.calculate_network_fee(
            transaction.Transaction._serializable_init()
        )
        self.assertEqual(123, response)

    async def test_find_states(self):
        key1 = b"\x0c\x00\x00\x00\x01"
        key2 = b"\x0c\x00\x00\x00\x02"
        key3 = b"\x0c\x00\x00\x00\x03"

        value1 = b"\x97\"\x8dq\xd20\xaf\xde\\\xce\x8f\xf9'\x1f*\x9d(\x88u\xf0"
        value2 = b"\x92,\x15\xa9\xa0\xe9\x00\x02\xed\xb4o\x1e>\xe4\xb7V\x8c\xb7%F"
        value3 = b"\xe0\x98^\x9d\xf0w\xb0\x88v\x1eV\xb3m\x97\xef\x89\x08F\x12\x13"

        captured1 = {
            "truncated": True,
            "next": 2,
            "results": [
                {
                    "key": base64.b64encode(key1).decode(),
                    "value": base64.b64encode(value1).decode(),
                },
                {
                    "key": base64.b64encode(key2).decode(),
                    "value": base64.b64encode(value2).decode(),
                },
            ],
        }
        captured2 = {
            "truncated": False,
            "next": 3,
            "results": [
                {
                    "key": base64.b64encode(key3).decode(),
                    "value": base64.b64encode(value3).decode(),
                }
            ],
        }
        self.mock_response(captured1)
        self.mock_response(captured2)
        from neo3.contracts.contract import CONTRACT_HASHES

        results = []
        async for k, v in self.client.find_states(CONTRACT_HASHES.MANAGEMENT, b"\x0c"):
            results.append((k, v))
        self.assertEqual(3, len(results))
        self.assertEqual(key1, results[0][0])
        self.assertEqual(value1, results[0][1])
        self.assertEqual(key2, results[1][0])
        self.assertEqual(value2, results[1][1])
        self.assertEqual(key3, results[2][0])
        self.assertEqual(value3, results[2][1])

    async def test_get_application_log_transaction(self):
        captured = {
            "txid": "0x7da6ae7ff9d0b7af3d32f3a2feb2aa96c2a27ef8b651f9a132cfaad6ef20724c",
            "executions": [
                {
                    "trigger": "Application",
                    "vmstate": "HALT",
                    "exception": None,
                    "gasconsumed": "9999540",
                    "stack": [],
                    "notifications": [
                        {
                            "contract": "0x70e2301955bf1e74cbb31d18c2f96972abadb328",
                            "eventname": "Transfer",
                            "state": {
                                "type": "Array",
                                "value": [
                                    {
                                        "type": "ByteString",
                                        "value": "4rZTInKT6ZxPKQbVNVOrtKZy34Y=",
                                    },
                                    {
                                        "type": "ByteString",
                                        "value": "+on7LBTfD1nd3wT25WUX8rNKrus=",
                                    },
                                    {"type": "Integer", "value": "10000000000"},
                                ],
                            },
                        }
                    ],
                }
            ],
        }
        self.mock_response(captured)
        tx_hash = types.UInt256.from_string(
            "0x7da6ae7ff9d0b7af3d32f3a2feb2aa96c2a27ef8b651f9a132cfaad6ef20724c"[2:]
        )
        response = await self.client.get_application_log_transaction(tx_hash)
        self.assertEqual(0, len(response.execution.stack))
        self.assertEqual("HALT", response.execution.state)
        self.assertEqual(1, len(response.execution.notifications))
        self.assertEqual("Transfer", response.execution.notifications[0].event_name)

    async def test_get_application_log_block(self):
        captured = {
            "blockhash": "0x577ee5cf7c589f608937287f11da965c0462a8fae77f29959c834cbce38cacac",
            "executions": [
                {
                    "trigger": "OnPersist",
                    "vmstate": "HALT",
                    "gasconsumed": "0",
                    "stack": [],
                    "notifications": [],
                },
                {
                    "trigger": "PostPersist",
                    "vmstate": "HALT",
                    "gasconsumed": "0",
                    "stack": [],
                    "notifications": [
                        {
                            "contract": "0xd2a4cff31913016155e38e474a2c06d08be276cf",
                            "eventname": "Transfer",
                            "state": {
                                "type": "Array",
                                "value": [
                                    {"type": "Any"},
                                    {
                                        "type": "ByteString",
                                        "value": "gSvnpwgsmEL8Ks76QRi4vFpbkYs=",
                                    },
                                    {"type": "Integer", "value": "50000000"},
                                ],
                            },
                        }
                    ],
                },
            ],
        }
        self.mock_response(captured)
        block_hash = types.UInt256.from_string(
            "0x577ee5cf7c589f608937287f11da965c0462a8fae77f29959c834cbce38cacac"[2:]
        )
        response = await self.client.get_application_log_block(block_hash)
        self.assertEqual(2, len(response.executions))
        self.assertEqual(0, len(response.executions[0].stack))
        self.assertEqual("HALT", response.executions[0].state)

        self.assertEqual(0, len(response.executions[1].stack))
        self.assertEqual("HALT", response.executions[1].state)
        self.assertEqual(1, len(response.executions[1].notifications))
        self.assertEqual("Transfer", response.executions[1].notifications[0].event_name)

    async def test_get_best_block_hash(self):
        hash_ = "0xbee7a65279d6b31cc45445a7579d4c4a4e52d1edc13cc7ec7a41f7b1affdf0ab"
        self.mock_response(hash_)
        response = await self.client.get_best_block_hash()
        self.assertEqual(types.UInt256.from_string(hash_[2:]), response)

    async def test_get_block(self):
        block_ = block.Block._serializable_init()
        self.mock_response(base64.b64encode(block_.to_array()).decode())
        response_block = await self.client.get_block(1)
        self.assertEqual(block_, response_block)

    async def test_get_block_count(self):
        count = 1
        self.mock_response(count)
        response = await self.client.get_block_count()
        self.assertEqual(count, response)

    async def test_get_block_hash(self):
        hash_ = "0xbee7a65279d6b31cc45445a7579d4c4a4e52d1edc13cc7ec7a41f7b1affdf0ab"
        self.mock_response(hash_)
        response = await self.client.get_block_hash(1)
        self.assertEqual(types.UInt256.from_string(hash_[2:]), response)

    async def test_get_block_header(self):
        block_ = block.Block._serializable_init()
        self.mock_response(base64.b64encode(block_.to_array()).decode())
        response_header = await self.client.get_block_header(1)
        self.assertEqual(block_.header, response_header)

    async def test_get_committee(self):
        points = [
            "02237309a0633ff930d51856db01d17c829a5b2e5cc2638e9c03b4cfa8e9c9f971",
            "022f1beae94cf0d266d7d26691b431958c8d13768103ab20aed817b57568da293f",
            "0239a37436652f41b3b802ca44cbcb7d65d3aa0b88c9a0380243bdbe1aaa5cb35b",
        ]
        self.mock_response(points)
        expected = tuple(
            map(
                lambda p: cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(p)),
                points,
            )
        )
        committee = await self.client.get_committee()
        self.assertEqual(expected, committee)

    async def test_get_connection_count(self):
        count = 1
        self.mock_response(count)
        response = await self.client.get_connection_count()
        self.assertEqual(count, response)

    async def test_get_contract_state(self):
        contract_hash_str = "0xb776afb6ad0c11565e70f8ee1dd898da43e51be1"
        contract_hash = types.UInt160.from_string(contract_hash_str[2:])
        captured = {
            "id": 1,
            "updatecounter": 0,
            "hash": contract_hash_str,
            "nef": {
                "magic": 860243278,
                "compiler": "Neo.Compiler.CSharp 3.0.0",
                "source": "",
                "tokens": [
                    {
                        "hash": "0xfffdc93764dbaddd97c48f252a53ea4643faa3fd",
                        "method": "update",
                        "paramcount": 3,
                        "hasreturnvalue": False,
                        "callflags": "All",
                    },
                    {
                        "hash": "0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5",
                        "method": "getCandidates",
                        "paramcount": 0,
                        "hasreturnvalue": True,
                        "callflags": "All",
                    },
                    {
                        "hash": "0x726cb6e0cd8628a1350a611384688911ab75f51b",
                        "method": "ripemd160",
                        "paramcount": 1,
                        "hasreturnvalue": True,
                        "callflags": "All",
                    },
                    {
                        "hash": "0x726cb6e0cd8628a1350a611384688911ab75f51b",
                        "method": "sha256",
                        "paramcount": 1,
                        "hasreturnvalue": True,
                        "callflags": "All",
                    },
                    {
                        "hash": "0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0",
                        "method": "serialize",
                        "paramcount": 1,
                        "hasreturnvalue": True,
                        "callflags": "All",
                    },
                    {
                        "hash": "0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0",
                        "method": "deserialize",
                        "paramcount": 1,
                        "hasreturnvalue": True,
                        "callflags": "All",
                    },
                ],
                "script": "NA5B+CfsjEBB+CfsjEBXAQAMCnN1cGVyQWRtaW40IHBoStgkA8oAFJcmEGhK2CQJSsoAFCgDOiIDWCICQFcAAXhBm/ZnzkGSXegxIgJAQZJd6DFAQZv2Z85AykBXAAEMCUZvcmJpZGRlbjSkQfgn7Iw0F3gMCnN1cGVyQWRtaW40EhHbICICQFcAAniqJgR5OkBXAAJ5eEGb9mfOQeY/GIRAQeY/GIRAVwADNVX///+qJhYMEU5vIGF1dGhvcml6YXRpb24uOnp5eDcAAEA3AABAVwABWXjbMItAi0DbMEBXCAoMCUZvcmJpZGRlbnhB+CfsjDSUNwEAcBDbIHFoSnLKcxB0Ih9qbM7BRXV2bTRhdwd4bweXJgoR2yBKcUUiCWycdGxrMOEMF1NlbmRlciBpcyBub3QgQ2FuZGlkYXRlaTVG////fwl/CH8Hfn18e3p5eBrASjRLcmo3BAB4NXP///80QhHbICICQDcBAEBXAgFaeNswi1uLcGjbKDcDADcCAHFpStgkCUrKABQoAzoiAkDbMEA3AgBANwMAQNsoQFcAAUBXAAJ5eEGb9mfOQeY/GIRAQeY/GIRANwQAQFcBAXg1Dv///zQXcGhK2CQDyhC3JghoNwUAIgULIgJAVwABeEGb9mfOQZJd6DEiAkBBkl3oMUA3BQBAVwMAWTQocBDEAHFoQZwI7ZwmF2hB81S/HXJqEc4LmCYHaWoRzs8i5WkiAkBXAAEaeEGb9mfOQd8wuJoiAkBB3zC4mkBBnAjtnEBB81S/HUDPQFcCAQwJRm9yYmlkZGVueEH4J+yMJgcR2yAiBzWY/f//NRv+//94NV/+//9weDVY/v//NWH///9xaUrYJAPKELcmCmg0DRHbICIHENsgIgJAVwABeNsoQZv2Z85BL1jF7UBBL1jF7UDPQFYEDBTARUMMYSJWDL3Fhow6TOAvAt28wWAMAgwh2zBiDAVBVuezJ9swYwwBd9swYUA=",
                "checksum": 3038242458,
            },
            "manifest": {
                "name": "CommitteeInfoContract",
                "groups": [],
                "features": {},
                "supportedstandards": [],
                "abi": {
                    "methods": [
                        {
                            "name": "verify",
                            "parameters": [],
                            "returntype": "Boolean",
                            "offset": 0,
                            "safe": False,
                        },
                        {
                            "name": "getAdmin",
                            "parameters": [],
                            "returntype": "Hash160",
                            "offset": 14,
                            "safe": False,
                        },
                        {
                            "name": "setAdmin",
                            "parameters": [{"name": "admin", "type": "Hash160"}],
                            "returntype": "Boolean",
                            "offset": 92,
                            "safe": False,
                        },
                        {
                            "name": "update",
                            "parameters": [
                                {"name": "nefFile", "type": "ByteArray"},
                                {"name": "manifest", "type": "String"},
                                {"name": "data", "type": "Any"},
                            ],
                            "returntype": "Void",
                            "offset": 168,
                            "safe": False,
                        },
                        {
                            "name": "setInfo",
                            "parameters": [
                                {"name": "sender", "type": "Hash160"},
                                {"name": "name", "type": "String"},
                                {"name": "location", "type": "String"},
                                {"name": "website", "type": "String"},
                                {"name": "email", "type": "String"},
                                {"name": "github", "type": "String"},
                                {"name": "telegram", "type": "String"},
                                {"name": "twitter", "type": "String"},
                                {"name": "description", "type": "String"},
                                {"name": "logo", "type": "String"},
                            ],
                            "returntype": "Boolean",
                            "offset": 224,
                            "safe": False,
                        },
                        {
                            "name": "getInfo",
                            "parameters": [{"name": "candidate", "type": "Hash160"}],
                            "returntype": "Any",
                            "offset": 448,
                            "safe": False,
                        },
                        {
                            "name": "getAllInfo",
                            "parameters": [],
                            "returntype": "Array",
                            "offset": 507,
                            "safe": False,
                        },
                        {
                            "name": "deleteInfo",
                            "parameters": [{"name": "candidate", "type": "Hash160"}],
                            "returntype": "Boolean",
                            "offset": 589,
                            "safe": False,
                        },
                        {
                            "name": "_initialize",
                            "parameters": [],
                            "returntype": "Void",
                            "offset": 694,
                            "safe": False,
                        },
                    ],
                    "events": [],
                },
                "permissions": [
                    {
                        "contract": "0x726cb6e0cd8628a1350a611384688911ab75f51b",
                        "methods": ["ripemd160", "sha256"],
                    },
                    {
                        "contract": "0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0",
                        "methods": ["deserialize", "serialize"],
                    },
                    {
                        "contract": "0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5",
                        "methods": ["getCandidates"],
                    },
                    {
                        "contract": "0xfffdc93764dbaddd97c48f252a53ea4643faa3fd",
                        "methods": ["update"],
                    },
                ],
                "trusts": [],
                "extra": {
                    "Author": "NEO",
                    "Email": "developer@neo.org",
                    "Description": "This is a Neo3 Contract",
                },
            },
        }
        self.mock_response(captured)

        response_state = await self.client.get_contract_state(contract_hash)
        self.assertEqual(contract_hash, response_state.hash)
        self.assertEqual("CommitteeInfoContract", response_state.manifest.name)

    async def test_get_nep17_balances(self):
        captured = {
            "balance": [
                {
                    "assethash": "0x70e2301955bf1e74cbb31d18c2f96972abadb328",
                    "amount": "3000000100000000",
                    "lastupdatedblock": 2,
                },
                {
                    "assethash": "0xf61eebf573ea36593fd43aa150c055ad7906ab83",
                    "amount": "99999900",
                    "lastupdatedblock": 2,
                },
            ],
            "address": "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF",
        }
        self.mock_response(captured)

        response = await self.client.get_nep17_balances("bogus_addr")
        self.assertEqual(2, len(response.balances))
        self.assertEqual(99999900, response.balances[1].amount)

    async def test_get_nep17_transfers(self):
        captured = {
            "sent": [],
            "received": [
                {
                    "timestamp": 1612690497725,
                    "assethash": "0xf61eebf573ea36593fd43aa150c055ad7906ab83",
                    "transferaddress": "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF",
                    "amount": "100",
                    "blockindex": 2,
                    "transfernotifyindex": 1,
                    "txhash": "0x5f957960a782514d6587c445288ee1cca7d6b0f952edc204f14d9be83b8152ff",
                },
                {
                    "timestamp": 1612690513541,
                    "assethash": "0x70e2301955bf1e74cbb31d18c2f96972abadb328",
                    "transferaddress": "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF",
                    "amount": "10000000000",
                    "blockindex": 3,
                    "transfernotifyindex": 0,
                    "txhash": "0xe42108b343626035cb51fbcb54949bb38aac50c8ba278841d304e9fdce0807ac",
                },
            ],
            "address": "NikhQp1aAD1YFCiwknhM5LQQebj4464bCJ",
        }

        self.mock_response(captured)

        response = await self.client.get_nep17_transfers("bogus_addr")
        self.assertEqual(2, len(response.received))
        self.assertEqual(100, response.received[0].amount)

    async def test_get_raw_mempool(self):
        tx1_hash = "0x0c65fbfd2598aee5f30cd18f1264b458f1db137c4a460f4a174facb3f2d59d06"
        tx2_hash = "0xc8040c285aa495f5b5e5b3761fd9333899f4ed902951c46d86c3bbb1cb12f2c0"

        resp = {"height": 5882071, "verified": [tx1_hash, tx2_hash], "unverified": []}
        self.mock_response(resp)

        expected_verified = [
            types.UInt256.from_string(tx1_hash[2:]),
            types.UInt256.from_string(tx2_hash[2:]),
        ]
        response = await self.client.get_raw_mempool()
        self.assertEqual(expected_verified, response.verified)

    async def test_get_next_block_validators(self):
        votes = 660174
        pub_key_raw = (
            "02237309a0633ff930d51856db01d17c829a5b2e5cc2638e9c03b4cfa8e9c9f971"
        )
        public_key = cryptography.ECPoint.deserialize_from_bytes(
            bytes.fromhex(pub_key_raw)
        )

        resp = [{"publickey": pub_key_raw, "votes": str(votes), "active": False}]
        self.mock_response(resp)

        response = await self.client.get_next_blockvalidators()
        self.assertEqual(1, len(response.validators))
        self.assertEqual(votes, response.validators[0].votes)
        self.assertEqual(public_key, response.validators[0].public_key)

    async def test_get_peers(self):
        peer1 = {"address": "47.90.28.99", "port": 21333}
        peer2 = {"address": "47.91.28.99", "port": 22333}

        resp = {"unconnected": [], "bad": [], "connected": [peer1, peer2]}
        self.mock_response(resp)

        response = await self.client.get_peers()
        self.assertEqual(2, len(response.connected))
        self.assertEqual(peer1["address"], response.connected[0].address)
        self.assertEqual(peer1["port"], response.connected[0].port)
        self.assertEqual(peer2["address"], response.connected[1].address)
        self.assertEqual(peer2["port"], response.connected[1].port)

    async def test_get_storage(self):
        storage_value = b"\x01\x02\x03"

        self.mock_response(base64.b64encode(storage_value).decode())

        dummy_contract_hash = types.UInt160.from_string(
            "0x99042d380f2b754175717bb932a911bc0bb0ad7d"[2:]
        )
        response = await self.client.get_storage(dummy_contract_hash, key=b"\x11")
        self.assertEqual(storage_value, response)

    async def test_get_transaction(self):
        tx = transaction.Transaction._serializable_init()
        # set some properties to ensure the TX will pass the deserialization checks
        tx.signers.append(verification.Signer._serializable_init())
        tx.script = b"\x01"
        tx.witnesses = [verification.Witness._serializable_init()]

        self.mock_response(base64.b64encode(tx.to_array()).decode())

        response_tx = await self.client.get_transaction(tx.hash())
        self.assertEqual(tx, response_tx)

    async def test_get_tx_height(self):
        height = 1
        tx_hash = "0c65fbfd2598aee5f30cd18f1264b458f1db137c4a460f4a174facb3f2d59d06"
        self.mock_response(height)
        response = await self.client.get_transaction_height(tx_hash)
        self.assertEqual(height, response)

    async def test_get_unclaimed_gas_(self):
        addr = "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF"
        value = 100
        self.mock_response({"unclaimed": str(value)})
        response_value = await self.client.get_unclaimed_gas(addr)
        self.assertEqual(value, response_value)

    async def test_get_version(self):
        user_agent = "/Neo:3.0.3/"
        captured = {
            "tcpport": 10333,
            "wsport": 10334,
            "nonce": 1930156121,
            "useragent": user_agent,
            "protocol": {
                "addressversion": 53,
                "network": 860833102,
                "validatorscount": 7,
                "msperblock": 15000,
                "maxtraceableblocks": 2102400,
                "maxvaliduntilblockincrement": 5760,
                "maxtransactionsperblock": 512,
                "memorypoolmaxtransactions": 50000,
                "initialgasdistribution": 5200000000000000,
            },
        }
        self.mock_response(captured)
        response = await self.client.get_version()
        self.assertEqual(captured["tcpport"], response.tcp_port)
        self.assertEqual(user_agent, response.user_agent)
        self.assertEqual(captured["protocol"]["network"], response.protocol.network)

    async def test_invoke_contract_verify(self):
        captured = {
            "script": "VgEMFFbIjRQK0swPKQN90Qp/AGCitShcYEBXAANAQZv2Z84MBWhlbGxvDAV3b3JsZFNB5j8YhEBXAQAMFFbIjRQK0swPKQN90Qp/AGCitShcQfgn7IxwaEA=",
            "state": "HALT",
            "gasconsumed": "1017810",
            "exception": None,
            "stack": [{"type": "Boolean", "value": True}],
        }

        self.mock_response(captured)

        contract_hash = "92f5c79b88560584a900cfec15b0e00dc4d58b59"
        response = await self.client.invoke_contract_verify(contract_hash)
        self.assertEqual(1, len(response.stack))
        self.assertTrue(response.stack[0].value)

    async def test_invoke_function(self):
        captured = {
            "script": "VgEMFFbIjRQK0swPKQN90Qp/AGCitShcYEBXAANAQZv2Z84MBWhlbGxvDAV3b3JsZFNB5j8YhEBXAQAMFFbIjRQK0swPKQN90Qp/AGCitShcQfgn7IxwaEA=",
            "state": "HALT",
            "gasconsumed": "1017810",
            "exception": None,
            "notifications": [],
            "stack": [
                {"type": "Boolean", "value": True},
                {
                    "type": "Map",
                    "value": [
                        {
                            "key": {"type": "ByteString", "value": "bmFtZQ=="},
                            "value": {
                                "type": "ByteString",
                                "value": "TyAjOTU3IEludGVyb3BlcmFiaWxpdHk=",
                            },
                        },
                        {
                            "key": {"type": "ByteString", "value": "b3duZXI="},
                            "value": {
                                "type": "ByteString",
                                "value": "/d5rikUuvYmC6NRzIP3I3ETAoOw=",
                            },
                        },
                        {
                            "key": {"type": "ByteString", "value": "bnVtYmVy"},
                            "value": {"type": "Integer", "value": "957"},
                        },
                        {
                            "key": {"type": "ByteString", "value": "aW1hZ2U="},
                            "value": {
                                "type": "ByteString",
                                "value": "aHR0cHM6Ly9uZW8ub3JnL0ludGVyb3BlcmFiaWxpdHkucG5n",
                            },
                        },
                        {
                            "key": {"type": "ByteString", "value": "dmlkZW8="},
                            "value": {"type": "Any"},
                        },
                    ],
                },
            ],
        }

        self.mock_response(captured)

        contract_hash = "92f5c79b88560584a900cfec15b0e00dc4d58b59"
        response = await self.client.invoke_function(contract_hash, "dummy_func")
        self.assertEqual(2, len(response.stack))
        self.assertTrue(response.stack[0].value)
        self.assertEqual(api.StackItemType.MAP, response.stack[1].type)
        self.assertEqual(5, len(response.stack[1].value))

    async def test_invoke_script(self):
        captured = {
            "script": "VgEMFFbIjRQK0swPKQN90Qp/AGCitShcYEBXAANAQZv2Z84MBWhlbGxvDAV3b3JsZFNB5j8YhEBXAQAMFFbIjRQK0swPKQN90Qp/AGCitShcQfgn7IxwaEA=",
            "state": "HALT",
            "gasconsumed": "1017810",
            "exception": None,
            "stack": [{"type": "Boolean", "value": True}],
            "notifications": [],
        }

        self.mock_response(captured)

        bogus_script = b"\x01"
        response = await self.client.invoke_script(bogus_script)
        self.assertEqual(1, len(response.stack))
        self.assertTrue(response.stack[0].value)

    async def test_send_tx(self):
        tx = transaction.Transaction._serializable_init()
        hash_ = "0x13ccdb9f7eda95a24aa5a4841b24fed957fe7f1b944996cbc2e92a4fa4f1fa73"
        hash_type = types.UInt256.from_string(hash_[2:])
        bogus_response = {"hash": hash_}
        self.mock_response(bogus_response)
        response_tx_id = await self.client.send_transaction(tx)
        self.assertEqual(hash_type, response_tx_id)

    async def test_send_block(self):
        block_ = block.Block._serializable_init()
        hash_ = "0x13ccdb9f7eda95a24aa5a4841b24fed957fe7f1b944996cbc2e92a4fa4f1fa73"
        hash_type = types.UInt256.from_string(hash_[2:])
        bogus_response = {"hash": hash_}
        self.mock_response(bogus_response)
        response_block_id = await self.client.send_block(block_)
        self.assertEqual(hash_type, response_block_id)

    async def test_validate_address(self):
        addr = "NPvKVTGZapmFWABLsyvfreuqn73jCjJtN1"
        resp = {"address": addr, "isvalid": True}
        self.mock_response(resp)
        response = await self.client.validate_address(addr)
        self.assertTrue(response)

    async def test_exception(self):
        with self.assertRaises(api.JsonRpcError) as context:
            error_response = {
                "jsonrpc": "2.0",
                "id": 1,
                "error": {
                    "code": -2146233086,
                    "message": "bogus_message",
                    "data": "bogus_data",
                },
            }
            self.helper.post("localhost", payload=error_response)
            await self.client.validate_address("abc")
        self.assertIn("-2146233086", str(context.exception))
        self.assertIn("bogus_message", str(context.exception))
        self.assertIn("bogus_data", str(context.exception))

    async def test_findstorage_issue_neogo(self):
        # test workaround for https://github.com/nspcc-dev/neo-go/issues/3370
        captured = {"results": None, "next": 0, "truncated": False}
        self.mock_response(captured)

        x = [(k, v) async for k, v in self.client.find_states(types.UInt160.zero())]
        self.assertEqual(0, len(x))


class TestStackItem(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.si_int = api.StackItem(api.StackItemType.INTEGER, 1)
        cls.si_bool = api.StackItem(api.StackItemType.BOOL, False)

    def test_as_bool(self):
        si = api.StackItem(api.StackItemType.BOOL, True)
        value = si.as_bool()

        self.assertIsInstance(value, bool)
        self.assertTrue(value)

        with self.assertRaises(ValueError) as context:
            self.si_int.as_bool()
        self.assertEqual(
            "item is not of type 'StackItemType.BOOL' but of type 'StackItemType.INTEGER'",
            str(context.exception),
        )

    def test_as_int(self):
        si = api.StackItem(api.StackItemType.INTEGER, 123)
        value = si.as_int()

        self.assertIsInstance(value, int)
        self.assertTrue(value)

        with self.assertRaises(ValueError) as context:
            self.si_bool.as_int()
        self.assertEqual(
            "item is not of type 'StackItemType.INTEGER' but of type 'StackItemType.BOOL'",
            str(context.exception),
        )

    def test_as_str(self):
        si = api.StackItem(api.StackItemType.BYTE_STRING, b"NEO")
        value = si.as_str()

        self.assertIsInstance(value, str)
        self.assertEqual("NEO", value)

        with self.assertRaises(ValueError) as context:
            self.si_int.as_str()
        self.assertEqual(
            "item is not of type 'StackItemType.BYTE_STRING' or 'StackItemType.BUFFER' but of type 'StackItemType.INTEGER'",
            str(context.exception),
        )

    def test_as_uint160(self):
        si = api.StackItem(api.StackItemType.BYTE_STRING, b"\x01" * 20)
        value = si.as_uint160()

        self.assertIsInstance(value, types.UInt160)

        with self.assertRaises(ValueError) as context:
            self.si_int.as_uint160()
        self.assertEqual(
            "item is not of type 'StackItemType.BYTE_STRING' or 'StackItemType.BUFFER' but of type 'StackItemType.INTEGER'",
            str(context.exception),
        )

    def test_as_uint256(self):
        si = api.StackItem(api.StackItemType.BYTE_STRING, b"\x01" * 32)
        value = si.as_uint256()

        self.assertIsInstance(value, types.UInt256)

        with self.assertRaises(ValueError) as context:
            self.si_int.as_uint256()
        self.assertEqual(
            "item is not of type 'StackItemType.BYTE_STRING' or 'StackItemType.BUFFER' but of type 'StackItemType.INTEGER'",
            str(context.exception),
        )

    def test_as_address(self):
        addr = "NSEB78A7DYhPG7cFT7x4YFhSKumBBm7RYk"
        script_hash = utils.address_to_script_hash(addr)

        si = api.StackItem(api.StackItemType.BYTE_STRING, script_hash.to_array())
        value = si.as_uint160()

        self.assertIsInstance(value, types.UInt160)
        self.assertEqual(script_hash, value)

        with self.assertRaises(ValueError) as context:
            self.si_int.as_address()
        self.assertEqual(
            "item is not of type 'StackItemType.BYTE_STRING' or 'StackItemType.BUFFER' but of type 'StackItemType.INTEGER'",
            str(context.exception),
        )

    def test_as_public_key(self):
        si = api.StackItem(
            api.StackItemType.BYTE_STRING,
            bytes.fromhex(
                "03b209fd4f53a7170ea4444e0cb0a6bb6a53c2bd016926989cf85f9b0fba17a70c"
            ),
        )
        value = si.as_public_key()

        self.assertIsInstance(value, cryptography.ECPoint)

        with self.assertRaises(ValueError) as context:
            self.si_int.as_public_key()
        self.assertEqual(
            "item is not of type 'StackItemType.BYTE_STRING' or 'StackItemType.BUFFER' but of type 'StackItemType.INTEGER'",
            str(context.exception),
        )

        # invalid public key
        si = api.StackItem(api.StackItemType.BYTE_STRING, b"\x01" * 64)
        with self.assertRaises(neo3crypto.ECCException) as context:
            value = si.as_public_key()
        self.assertEqual("Failed public key validation", str(context.exception))



from typing import TypeAlias

# Note that a NeoAddress is just a base58check encoded (address version + script hash).
# * The address version is a fixed value since the inception of the chain.
# * The script hash is the public key of an account (ECPair) wrapped with some extra data and hashed
#   with ripemd160. It is represented in the code with the UInt160 type from the `core` package
NeoAddress: TypeAlias = str



from __future__ import annotations
import abc
from enum import IntEnum
from neo3.core import types, Size as s, utils, serialization
from neo3.network.payloads import verification
from collections.abc import Sequence


class InventoryType(IntEnum):
    TX = 0x2B
    BLOCK = 0x2C
    CONSENSUS = 0x2D
    EXTENSIBLE = 0x2E


class InventoryPayload(serialization.ISerializable):
    """
    A payload used to share inventory hashes.

    Use by `getdata`, `getblocks` and `mempool` message types.
    """

    def __init__(self, type: InventoryType, hashes: Sequence[types.UInt256]):
        """
        Create payload.

        Args:
            type: indicator to what type of object the hashes of this payload relate to.
            hashes: hashes of "type" objects.
        """
        self.type = type
        self.hashes = hashes

    def __len__(self):
        """Get the total size in bytes of the object."""
        return s.uint8 + utils.get_var_size(self.hashes)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.type)
        writer.write_var_int(len(self.hashes))
        for h in self.hashes:  # type: types.UInt256
            writer.write_bytes(h.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.type = InventoryType(reader.read_uint8())
        self.hashes = reader.read_serializable_list(types.UInt256)

    @classmethod
    def _serializable_init(cls):
        return cls(InventoryType.BLOCK, [])


class IInventory(verification.IVerifiable):
    """
    Inventory interface.
    """

    @abc.abstractmethod
    def hash(self) -> types.UInt256:
        """"""

    @property
    @abc.abstractmethod
    def inventory_type(self) -> InventoryType:
        """"""



"""
NEP-15 contract manifest classes for describing smart contract access control.
"""
from __future__ import annotations
import base64
import binascii
import orjson as json  # type: ignore
from typing import Callable, Optional, Any
from neo3.contracts import utils as contractutils, abi
from neo3.core import serialization, types, cryptography, utils as coreutils, interfaces
from neo3.core.serialization import BinaryReader, BinaryWriter


class ContractGroup(interfaces.IJson):
    """
    Describes a set of mutually trusted contracts.

    See Also:
        ContractManifest.
    """

    def __init__(self, public_key: cryptography.ECPoint, signature: bytes):
        self.public_key = public_key
        self.signature = signature

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.public_key == other.public_key and self.signature == other.signature

    def is_valid(self, contract_hash: types.UInt160) -> bool:
        """
        Validate if the group has agreed on allowing the specific contract_hash.

        Args:
            contract_hash:
        """
        return cryptography.verify_signature(
            contract_hash.to_array(),
            self.signature,
            self.public_key.encode_point(False),
            cryptography.ECCCurve.SECP256R1,
        )

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "pubkey": str(self.public_key),
            "signature": base64.b64encode(self.signature).decode(),
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractGroup:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the signature length is not 64.
        """
        pubkey = contractutils.validate_type(json["pubkey"], str)
        c = cls(
            public_key=cryptography.ECPoint.deserialize_from_bytes(
                binascii.unhexlify(pubkey)
            ),
            signature=base64.b64decode(
                contractutils.validate_type(json["signature"], str).encode("utf8")
            ),
        )
        if len(c.signature) != 64:
            raise ValueError("Format error - invalid signature length")
        return c


class ContractPermission(interfaces.IJson):
    """
    Describes a single set of outgoing call restrictions for a `System.Contract.Call` SYSCALL.
    It describes what other smart contracts the executing contract is allowed to call and what exact methods on the
    other contract are allowed to be called. This is enforced during runtime.

    Example:
        Contract A (the executing contract) wants to call method `x` on Contract B. The runtime will query the manifest
        of Contract A and ask if this is allowed. The Manifest will search through its permissions (a list of
        ContractPermission objects) and ask if it `is_allowed()`.
    """

    def __init__(
        self,
        contract: ContractPermissionDescriptor,
        methods: WildcardContainer,
    ):
        self.contract = contract
        self.methods = methods

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.contract == other.contract and self.methods == other.methods

    @classmethod
    def default_permissions(cls) -> ContractPermission:
        """
        Construct a ContractPermission which allows any contract and any method to be called.
        """
        return cls(
            ContractPermissionDescriptor(),  # with no parameters equals to Wildcard
            WildcardContainer.create_wildcard(),
        )

    def is_allowed(
        self,
        target_contract_hash: types.UInt160,
        target_manifest: ContractManifest,
        target_method: str,
    ) -> bool:
        """
        Return if it is allowed to call `target_method` on the target contract.

        Args:
            target_contract_hash: the contract hash of the contract to be called.
            target_manifest: the contract manifest of the contract to be called.
            target_method: the method of the contract to be called.
        """
        if self.contract.is_hash:
            if not self.contract.contract_hash == target_contract_hash:
                return False
        elif self.contract.is_group:
            results = list(
                map(
                    lambda p: p.public_key != self.contract.group,
                    target_manifest.groups,
                )
            )
            if all(results):
                return False
        return self.methods.is_wildcard or target_method in self.methods

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = self.contract.to_json()
        # because NEO C# returns a string from "method" instead of sticking to a standard interface
        json.update({"methods": self.methods.to_json()["wildcard"]})
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractPermission:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if a method is zero length.
        """
        cpd = ContractPermissionDescriptor.from_json(json)
        json_wildcard = {"wildcard": json["methods"]}
        methods = WildcardContainer.from_json(json_wildcard)
        for m in methods:
            if len(m) == 0:
                raise ValueError("Format error - methods cannot have length 0")
        return cls(cpd, methods)


class WildcardContainer(interfaces.IJson):
    """
    An internal helper class for ContractManifest attributes.
    """

    def __init__(self, data: Optional[list] = None):
        self._is_wildcard = False
        self._data = data if data else []

    def __contains__(self, item):
        return item in self._data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self._data == other._data

    @classmethod
    def create_wildcard(cls) -> WildcardContainer:
        """
        Creates an instance that indicates any value is allowed.
        """
        instance = cls()
        instance._is_wildcard = True
        return instance

    @property
    def is_wildcard(self) -> bool:
        """
        Indicates if the container is configured to allow all values.
        """
        return self._is_wildcard

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        if self.is_wildcard:
            return {"wildcard": "*"}
        return {"wildcard": list(map(lambda d: str(d), self._data))}

    @classmethod
    def from_json(cls, json: dict):
        """
        Parse object out of JSON data.

        Note: if the value is not '*', and is a Python list, then it will assume
        that the list members are strings or convertible via str().

        If the wildcard should contain other data types, use the alternative `from_json_as_type()` method.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        value = json.get("wildcard", None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot recreate wildcard from None")
        if value == "*":
            return WildcardContainer.create_wildcard()
        if isinstance(value, list):
            return WildcardContainer(data=list(map(lambda d: str(d), value)))
        raise ValueError(
            f"Invalid JSON - Cannot deduce WildcardContainer type from: {value}"
        )

    @classmethod
    def from_json_as_type(cls, json: dict, conversion_func: Callable):
        """
            Parse object out of JSON data.

            Note: if the value is not '*', and is a Python list, then it will use `conversion_func` to
            parse the members into the expected types.

        Args:
            json: a dictionary.
            conversion_func: a callable that takes 1 argument, which is the element in the value list.

            Example with UInt160:
                {'wildcard': ['0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01', 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa']}

                 the first call has as argument '0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01'.
                 to process this example call
                 WildcardContainer.from_json_as_type(json_data, lambda f: types.UInt160.from_string(f))

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        value = json.get("wildcard", None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot recreate wildcard from None")
        if value == "*":
            return WildcardContainer.create_wildcard()
        if isinstance(value, list):
            return WildcardContainer(
                data=list(map(lambda d: conversion_func(d), value))
            )
        raise ValueError(
            f"Invalid JSON - Cannot deduce WildcardContainer type from: {value}"
        )


class ContractPermissionDescriptor(interfaces.IJson):
    """
    A restriction object that limits the smart contract's calling abilities. Enforced at runtime.

    See Also: ContractManifest.
    """

    def __init__(
        self,
        contract_hash: Optional[types.UInt160] = None,
        group: Optional[cryptography.ECPoint] = None,
    ):
        """
        Create a contract hash or group based restriction. Mutually exclusive.
        Supply no arguments to create a wildcard permission descriptor.

        Raises:
            ValueError: if both contract hash and group arguments are supplied.
        """
        if contract_hash is not None and group is not None:
            raise ValueError("Maximum 1 argument can be supplied")
        self.contract_hash = contract_hash
        self.group = group

    def __eq__(self, other):
        return self.contract_hash == other.contract_hash and self.group == other.group

    @property
    def is_hash(self) -> bool:
        """
        Indicates if the permission is limited to a specific contract hash.
        """
        return self.contract_hash is not None

    @property
    def is_group(self) -> bool:
        """
        Indicates if the permission is limited to a specific group.
        """
        return self.group is not None

    @property
    def is_wildcard(self) -> bool:
        """
        Indicates if the permission is not limited to a specific contract or a specific group.
        """
        return not self.is_hash and not self.is_group

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        # NEO C# deviates here. They return a string
        if self.contract_hash:
            val = "0x" + str(self.contract_hash)
        elif self.group:
            val = str(self.group)
        else:
            val = "*"
        return {"contract": val}

    @classmethod
    def from_json(cls, json: dict) -> ContractPermissionDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        # catches both missing key and None as value
        value = json.get("contract", None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot deduce permission type from None")

        if len(value) == 42:
            return cls(contract_hash=types.UInt160.from_string(value[2:]))
        if len(value) == 66:
            ecpoint = cryptography.ECPoint.deserialize_from_bytes(
                binascii.unhexlify(value)
            )
            return cls(group=ecpoint)
        if value == "*":
            return cls()  # no args == wildcard
        raise ValueError(f"Invalid JSON - Cannot deduce permission type from: {value}")

    def to_array(self) -> bytes:
        """Serialize the object."""
        if self.is_hash:
            return self.contract_hash.to_array()  # type: ignore
        if self.is_group:
            return self.group.to_array()  # type: ignore
        # wildcard
        return b""


class ContractManifest(serialization.ISerializable, interfaces.IJson):
    """
    A description of a smart contract's abilities (callable methods & events) as well as a set of restrictions
    describing what external contracts and methods are allowed to be called.

    For more information see:
    https://github.com/neo-project/proposals/blob/3e492ad05d9de97abb6524fb9a73714e2cdc5461/nep-15.mediawiki
    """

    #: The maximum byte size after serialization to be considered valid a valid contract.
    MAX_LENGTH = 0xFFFF

    def __init__(self, name: Optional[str] = None):
        #: Contract name
        self.name: str = name if name else ""
        #: A group represents a set of mutually trusted contracts. A contract will trust and allow any contract in the
        #: same group to invoke it.
        self.groups: list[ContractGroup] = []

        #: The list of NEP standards supported e.g. "NEP-3"
        self.supported_standards: list[str] = []

        #: For technical details of ABI, please refer to NEP-14: NeoContract ABI.
        #: https://github.com/neo-project/proposals/blob/d1f4e9e1a67d22a5755c45595121f80b0971ea64/nep-14.mediawiki
        self.abi: abi.ContractABI = abi.ContractABI(events=[], methods=[])

        #: Permissions describe what external contract(s) and what method(s) on these are allowed to be invoked.
        self.permissions: list[ContractPermission] = [
            ContractPermission.default_permissions()
        ]

        # Update trusts/safe_methods with outcome of https://github.com/neo-project/neo/issues/1664
        # Unfortunately we have to add this nonsense logic or we get deviating VM results.
        self.trusts = WildcardContainer()  # for UInt160 types

        #: Optional user defined data
        self.extra: Optional[dict] = None

    def __len__(self):
        return coreutils.get_var_size(str(self.to_json()).replace(" ", ""))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            self.name == other.name
            and self.groups == other.groups
            and self.abi == other.abi
            and self.permissions == other.permissions
            and self.trusts == other.trusts
            and self.extra == other.extra
        )

    def __str__(self):
        return json.dumps(self.to_json()).decode()

    def serialize(self, writer: BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_string(str(self))

    def deserialize(self, reader: BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self._deserialize_from_json(json.loads(reader.read_var_string(self.MAX_LENGTH)))

    def _deserialize_from_json(self, json: dict) -> None:
        if json["name"] is None:
            self.name = ""
        else:
            self.name = contractutils.validate_type(json["name"], str)
        self.abi = abi.ContractABI.from_json(json["abi"])
        self.groups = list(map(lambda g: ContractGroup.from_json(g), json["groups"]))

        if len(json["features"]) != 0:
            raise ValueError(
                "Manifest features is reserved and cannot have any content at this time"
            )

        self.supported_standards = list(
            map(
                lambda ss: contractutils.validate_type(ss, str),
                json["supportedstandards"],
            )
        )
        self.permissions = list(
            map(lambda p: ContractPermission.from_json(p), json["permissions"])
        )

        if json["trusts"] == "*":
            self.trusts = WildcardContainer.create_wildcard()
        else:
            self.trusts = WildcardContainer.from_json_as_type(
                {"wildcard": json["trusts"]},
                lambda t: ContractPermissionDescriptor.from_json({"contract": t}),
            )

        # converting json key/value back to default WildcardContainer format
        self.extra = json["extra"]

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        if self.trusts.is_wildcard:
            trusts = "*"
        else:
            trusts = list(map(lambda m: m.to_json()["contract"], self.trusts))  # type: ignore
        json: dict[str, Any] = {
            "name": self.name if self.name else None,
            "groups": list(map(lambda g: g.to_json(), self.groups)),
            "features": {},
            "supportedstandards": self.supported_standards,
            "abi": self.abi.to_json(),
            "permissions": list(map(lambda p: p.to_json(), self.permissions)),
            "trusts": trusts,
            "extra": self.extra,
        }
        return json

    @classmethod
    def from_json(cls, json: dict):
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raise:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the manifest name property has an incorrect format.
            ValueError: if the manifest support standards contains an string.
        """
        manifest = cls()
        manifest._deserialize_from_json(json)
        if manifest.name is None or len(manifest.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        for s in manifest.supported_standards:
            if len(s) == 0:
                raise ValueError(
                    "Format error - supported standards cannot be zero length"
                )
        return manifest

    def is_valid(self, contract_hash: types.UInt160) -> bool:
        """
        Validates the if any of the manifest groups signed the requesting `contract_hash` as permissive.

        An example use-case is to allow creation and updating of smart contracts by a select group.

        Args:
            contract_hash: target contract hash.
        """
        result = list(map(lambda g: g.is_valid(contract_hash), self.groups))
        return all(result)

    def can_call(
        self,
        target_contract_hash: types.UInt160,
        target_manifest: ContractManifest,
        target_method: str,
    ) -> bool:
        """
        Check if this contract is allowed to call `target_method` on `target_contract`.
        """
        results = list(
            map(
                lambda p: p.is_allowed(
                    target_contract_hash, target_manifest, target_method
                ),
                self.permissions,
            )
        )
        return any(results)

    def contains_group(self, public_key: cryptography.ECPoint) -> bool:
        """
        Check if group exists.

        Args:
            public_key: needle to search for.

        Returns:
            `True` if found. `False` otherwise.
        """
        for g in self.groups:
            if public_key == g.public_key:
                return True
        return False

    @classmethod
    def from_file(cls, path: str):
        """
        Create object from a file.

        Args:
            path: location of the file.

        Raises:
            FileNotFoundError: if the path is invalid.
            ValueError: if the file is not a valid ContractManifest.
        """
        with open(path, "rb") as f:
            manifest_bytes = f.read()
            manifest_json = json.loads(manifest_bytes.decode("utf-8"))
            try:
                return cls.from_json(manifest_json)
            except ValueError as e:
                raise ValueError(f"Failed manifest validation with: {e}")

    @classmethod
    def _serializable_init(cls):
        return cls()



import unittest
from neo3.core import types
from neo3.api.wrappers import _check_address_and_convert


class WrapperUtilsTest(unittest.TestCase):
    def test_check_address_and_convert(self):
        hash_in = types.UInt160.from_string(
            "0x7e9237a93f64407141a5b86c760200c66c81e2ec"
        )
        self.assertIsInstance(_check_address_and_convert(hash_in), types.UInt160)

        with self.assertRaises(ValueError) as context:
            _check_address_and_convert(object())
        self.assertEqual(
            "Input is of type <class 'object'> expected UInt160 or NeoAddress(str)",
            str(context.exception),
        )

        invalid_address = "NgNJsBfhcoJSm6MVMpMeGLqEK5mSQXuJTt"
        with self.assertRaises(ValueError) as context:
            _check_address_and_convert(invalid_address)
        self.assertEqual("Invalid checksum", str(context.exception))

        valid_address = "NgNJsBfhcoJSm6MVMpMeGLqEK5mSQXuJTq"
        self.assertIsInstance(_check_address_and_convert(valid_address), types.UInt160)



import unittest
from neo3.contracts import contract, nef, manifest
from neo3.core import types


class ContractStateTestCase(unittest.TestCase):
    def test_equals(self):
        manifest_ = manifest.ContractManifest()
        nef_ = nef.NEF()
        state = contract.ContractState(1, nef_, manifest_, 0, types.UInt160.zero())
        clone = contract.ContractState(1, nef_, manifest_, 0, types.UInt160.zero())
        self.assertEqual(state, clone)

        nef2 = nef.NEF()
        state2 = contract.ContractState(
            2, nef2, manifest_, 0, types.UInt160(b"\x01" * 20)
        )
        self.assertNotEqual(state, state2)
        self.assertNotEqual(state, None)
        self.assertNotEqual(state, object())



"""
NEP-14 Application Binary Interface classes.
"""
from __future__ import annotations
from typing import Optional
from collections.abc import Sequence
from enum import IntEnum
from neo3.core import interfaces
from neo3.contracts import utils as contractutils


class ContractParameterType(IntEnum):
    """
    Type information for a contract method parameter.
    """

    ANY = 0x00
    BOOLEAN = 0x10
    INTEGER = 0x11
    BYTEARRAY = 0x12
    STRING = 0x13
    HASH160 = 0x14
    HASH256 = 0x15
    PUBLICKEY = 0x16
    SIGNATURE = 0x17
    ARRAY = 0x20
    MAP = 0x22
    INTEROPINTERFACE = 0x30
    VOID = 0xFF

    def PascalCase(self) -> str:
        if self == ContractParameterType.BYTEARRAY:
            return "ByteArray"
        elif self == ContractParameterType.INTEROPINTERFACE:
            return "InteropInterface"
        elif self == ContractParameterType.PUBLICKEY:
            return "PublicKey"
        else:
            return self.name.title()


class ContractParameterDefinition(interfaces.IJson):
    """
    A parameter description for a contract Method or Event.
    """

    def __init__(self, name: str, type: ContractParameterType):
        """
        Args:
            name: the human-readable identifier.
            type: the type of parameter.
        """
        self.name = name
        self.type = type

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.name == other.name and self.type == other.type

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {"name": self.name, "type": self.type.PascalCase()}
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractParameterDefinition:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the manifest name property has an incorrect format.
            ValueError: if the type is VOID.
        """
        c = cls(
            name=contractutils.validate_type(json["name"], str),
            type=ContractParameterType[
                contractutils.validate_type(json["type"], str).upper()
            ],
        )
        if c.name is None or len(c.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        if c.type == ContractParameterType.VOID:
            raise ValueError("Format error - parameter type VOID is not allowed")
        return c


class ContractEventDescriptor(interfaces.IJson):
    """
    A description for an event that a contract can broadcast.
    """

    def __init__(self, name: str, parameters: Sequence[ContractParameterDefinition]):
        """

        Args:
            name: the human-readable identifier of the event.
            parameters: the list of parameters the event takes.
        """
        self.name = name
        self.parameters = parameters

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.name == other.name and self.parameters == other.parameters

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "name": self.name,
            "parameters": list(map(lambda p: p.to_json(), self.parameters)),
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractEventDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the `name` property has an incorrect format.
        """
        c = cls(
            name=contractutils.validate_type(json["name"], str),
            parameters=list(
                map(
                    lambda p: ContractParameterDefinition.from_json(p),
                    json["parameters"],
                )
            ),
        )
        if c.name is None or len(c.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        return c


class ContractMethodDescriptor(ContractEventDescriptor, interfaces.IJson):
    """
    A description of a callable method on a contract.
    """

    def __init__(
        self,
        name: str,
        offset: int,
        parameters: Sequence[ContractParameterDefinition],
        return_type: ContractParameterType,
        safe: bool,
    ):
        """
        Args:
            name: the human-readable identifier of the method.
            offset: script offset.
            parameters: the list of parameters the method takes.
            return_type: the type of the returned value.
        """
        super(ContractMethodDescriptor, self).__init__(name, parameters)
        self.offset = offset
        self.return_type = return_type
        self.safe = safe

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            self.name == other.name
            and self.parameters == other.parameters
            and self.offset == other.offset
            and self.return_type == other.return_type
            and self.safe == other.safe
        )

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = super(ContractMethodDescriptor, self).to_json()
        json.update(
            {
                "returntype": self.return_type.PascalCase(),
                "offset": self.offset,
                "safe": self.safe,
            }
        )
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractMethodDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the manifest name property has an incorrect format.
            ValueError: if the offset is negative.
        """
        c = cls(
            name=contractutils.validate_type(json["name"], str),
            offset=contractutils.validate_type(json["offset"], int),
            parameters=list(
                map(
                    lambda p: ContractParameterDefinition.from_json(p),
                    json["parameters"],
                )
            ),
            return_type=ContractParameterType[
                contractutils.validate_type(json["returntype"], str).upper()
            ],
            safe=contractutils.validate_type(json["safe"], bool),
        )
        if c.name is None or len(c.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        if c.offset < 0:
            raise ValueError("Format error - negative offset not allowed")
        return c

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> {self.name}"


class ContractABI(interfaces.IJson):
    """
    The smart contract application binary interface describes the callable events and contracts for a given
     smart contract.
    """

    def __init__(
        self,
        methods: Sequence[ContractMethodDescriptor],
        events: Sequence[ContractEventDescriptor],
    ):
        """
        Args:
            methods: the available methods in the contract.
            events: the various events that can be broadcast by the contract.
        """
        self.methods = methods
        self.events = events

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.methods == other.methods and self.events == other.events

    def get_method(
        self, name, parameter_count: int
    ) -> Optional[ContractMethodDescriptor]:
        """
        Return the ContractMethodDescriptor matching the name (and optional parameter count) or `None` otherwise.

        Args:
            name: the name of the method to return.
            parameter_count: the expected number of parameters teh method has.
        """
        if parameter_count < -1 or parameter_count > 0xFFFF:
            raise ValueError("Parameter count is out of range")

        if parameter_count >= 0:
            for m in self.methods:
                if m.name == name and len(m.parameters) == parameter_count:
                    return m
            else:
                return None
        else:
            for m in self.methods:
                if m.name == name:
                    return m
            else:
                return None

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "methods": list(map(lambda m: m.to_json(), self.methods)),
            "events": list(map(lambda e: e.to_json(), self.events)),
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractABI:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the contract has no methods.
        """
        c = cls(
            methods=list(
                map(lambda m: ContractMethodDescriptor.from_json(m), json["methods"])
            ),
            events=list(
                map(lambda e: ContractEventDescriptor.from_json(e), json["events"])
            ),
        )
        if len(c.methods) == 0:
            raise ValueError("Invalid contract - contract has no methods")
        return c



import unittest
from neo3.contracts import utils, nef
from neo3.core import types


class TestContractUtils(unittest.TestCase):
    def test_get_contract_hash(self):
        nef_ = nef.NEF("test", b"\x01\x02\x03")
        actual = utils.get_contract_hash(types.UInt160.zero(), nef_.checksum, "test")
        expected = types.UInt160.from_string(
            "0x576c9c6f22eea8fd823155b00141a4327bac8263"
        )
        self.assertEqual(expected, actual)

        actual = utils.get_contract_hash(
            types.UInt160.from_string("0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01"),
            nef_.checksum,
            "test",
        )
        expected = types.UInt160.from_string(
            "0x55f776130883b2d486dec295ca74533663d0f8ea"
        )
        self.assertEqual(expected, actual)



import aiodns
import logging
import socket
import asyncio
from unittest import mock, IsolatedAsyncioTestCase, TestCase, skip
from neo3.network import node, message, capabilities
from neo3.network.convenience import nodemanager, requestinfo
from neo3.network.payloads import address, version, ping
from neo3 import network_logger
from neo3.settings import settings
from datetime import datetime
from copy import deepcopy
from neo3.core import msgrouter


class NodeManagerTestCase(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.nodemgr = nodemanager.NodeManager()

    def setUp(self) -> None:
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()
        settings.reset_settings_to_default()

        self.node1 = node.NeoNode(object(), object())
        self.node2 = node.NeoNode(object(), object())

        self.addr1 = address.NetworkAddress(address="127.0.0.1:10333")
        self.addr2 = address.NetworkAddress(address="127.0.0.2:20333")
        self.node1.address = self.addr1
        self.node2.address = self.addr2

    def test_get_node_addresses(self):
        self.nodemgr.nodes = [self.node1, self.node2]
        addr_list = self.nodemgr.get_node_addresses()
        self.assertEqual([self.addr1, self.addr2], addr_list)

    def test_get_node_by_id(self):
        self.nodemgr.nodes = [self.node1, self.node2]
        n = self.nodemgr.get_node_by_nodeid(self.node1.nodeid)
        self.assertEqual(self.node1, n)

        random_id = 999
        n = self.nodemgr.get_node_by_nodeid(random_id)
        self.assertEqual(None, n)

    def test_get_node_with_height1(self):
        # no connected nodes = early exit
        self.assertEqual(None, self.nodemgr.get_node_with_height(height=1))

    def test_get_node_with_height2(self):
        # can select from 2 , should get the one with the highest weight
        self.node1.best_height = 10
        # increasing error count lowers weight
        self.node1.nodeweight.error_response_count = 5

        self.node2.best_height = 10

        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        self.assertEqual(self.node2, self.nodemgr.get_node_with_height(height=1))

    def test_get_node_with_height3(self):
        # should return None if none of the nodes have the requested height
        self.node1.best_height = 0
        self.node2.best_height = 0
        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        self.assertEqual(None, self.nodemgr.get_node_with_height(height=10))

    def test_get_least_failed_node1(self):
        # should return None if none of the nodes have the requested height
        self.node1.best_height = 0
        self.node2.best_height = 0
        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        ri = requestinfo.RequestInfo(height=10)
        self.assertEqual(None, self.nodemgr.get_least_failed_node(ri))

    def test_get_least_failed_node2(self):
        # should return the node with the lowest fail count for the request
        # for this test we setup that node1 has failed before, whereas node2 hasn't
        self.node1.best_height = 10
        self.node2.best_height = 10

        ri = requestinfo.RequestInfo(height=1)
        # pretend node 1 has already failed once for this request
        ri.failed_nodes[self.node1.nodeid] = 1

        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        self.assertEqual(self.node2, self.nodemgr.get_least_failed_node(ri))

    def test_get_least_failed_node3(self):
        # should not return a node that is in the process of disconnecting but does have the correct height
        self.node1.best_height = 10
        self.node1.disconnecting = True
        self.nodemgr.nodes = [self.node1, self.node2]
        ri = requestinfo.RequestInfo(height=10)
        self.assertEqual(None, self.nodemgr.get_least_failed_node(ri))

    def test_increase_node_error_count(self):
        self.nodemgr.nodes = [self.node1, self.node2]

        # we expect self.assertLogs to raise an AssertionError because there will be no DEBUG output logs
        # we call self.assertLogs because we want to test for no DBEUG logs being produced
        with self.assertRaises(AssertionError) as assert_log_context:
            with self.assertLogs(network_logger, "DEBUG") as context:
                self.nodemgr.increase_node_error_count(self.node1.nodeid)
        self.assertEqual(1, self.node1.nodeweight.error_response_count)
        self.assertEqual(0, len(context.output))

    def test_increase_node_error_count2(self):
        self.nodemgr.nodes = [self.node1, self.node2]
        # set a current value that certainly will exceed whatever the user configuration will say
        self.node1.nodeweight.error_response_count += 999

        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch.object(
                self.node1,
                "disconnect",
                new_callable=mock.MagicMock,  # forcing this to non-async version to prevent runtime warning
            ):
                with mock.patch("asyncio.create_task") as mocked_create_task:
                    self.nodemgr.increase_node_error_count(self.node1.nodeid)
        self.assertEqual(1000, self.node1.nodeweight.error_response_count)
        self.assertIn("max error count threshold exceeded", context.output[0])

    def test_increase_node_timeout_count(self):
        self.nodemgr.nodes = [self.node1, self.node2]

        self.nodemgr.increase_node_timeout_count(self.node1.nodeid)
        self.assertEqual(1, self.node1.nodeweight.timeout_count)

    def test_increase_node_timeout_count2(self):
        # test we detect max count threshold exceeding and schedule a disconnect task
        self.nodemgr.nodes = [self.node1, self.node2]
        # set a current value that certainly will exceed whatever the user configuration will say
        self.node1.nodeweight.timeout_count += 999

        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch.object(
                self.node1,
                "disconnect",
                new_callable=mock.MagicMock,  # forcing this to non-async version to prevent runtime warning):
            ):
                with mock.patch("asyncio.create_task"):
                    self.nodemgr.increase_node_timeout_count(self.node1.nodeid)
        self.assertEqual(1000, self.node1.nodeweight.timeout_count)
        self.assertIn("max timeout count threshold exceeded", context.output[0])


caps = [capabilities.FullNodeCapability(0)]
m_version = message.Message(
    msg_type=message.MessageType.VERSION,
    payload=version.VersionPayload(
        nonce=123, user_agent="NEO3-MOCK-CLIENT", capabilities=caps
    ),
)
m_verack = message.Message(msg_type=message.MessageType.VERACK)

m_ping = message.Message(
    msg_type=message.MessageType.PING, payload=ping.PingPayload(height=0)
).to_array()


async def fake_create_task(coro):
    await coro


class NodeManagerTestCase2(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # stdio_handler = logging.StreamHandler()
        # stdio_handler.setLevel(logging.DEBUG)
        # stdio_handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s - %(module)s:%(lineno)s %(message)s"))
        async_logger = logging.getLogger("asyncio")
        # async_logger.addHandler(stdio_handler)
        async_logger.setLevel(logging.DEBUG)

        network_logger = logging.getLogger("neo3.network")
        # network_logger.addHandler(stdio_handler)
        network_logger.setLevel(logging.DEBUG)

        cls.nodemgr = nodemanager.NodeManager()

    def setUp(self) -> None:
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()
        settings.reset_settings_to_default()

    async def test_start_shutdown(self):
        settings.register({"network": {"seedlist": ["127.0.0.1:1111"], "magic": 769}})

        self.nodemgr.start()
        # need to keep this outside or it will go out of scope and close the socket
        r, w = socket.socketpair()

        def client_provider():
            loop = asyncio.get_running_loop()
            v = deepcopy(m_version)
            v.payload.magic = 769
            loop.call_soon(w.send, v.to_array())
            loop.call_soon(w.send, m_verack.to_array())
            yield r, w

        # configure nodemanager to use a mock client
        self.nodemgr._test_client_provider = client_provider
        self.nodemgr._test_data = {"peername": ("127.0.0.1", 1111)}

        call_back_result = None

        def client_connect_done(node_instance, failure):
            nonlocal call_back_result
            call_back_result = node_instance

        # listen to a connection done event
        msgrouter.on_client_connect_done += client_connect_done

        # initial state
        self.assertFalse(self.nodemgr.is_running)
        # advance loop
        await asyncio.sleep(0.1)
        self.assertTrue(self.nodemgr.is_running)
        # 4 looping service tasks
        self.assertEqual(4, len(self.nodemgr.tasks))
        self.assertIsInstance(call_back_result, node.NeoNode)
        await self.nodemgr.shutdown()
        for t in self.nodemgr.tasks:
            self.assertTrue(t.done())
        r.close()
        w.close()

    async def test_fill_open_connection_spots_no_clients(self):
        # create 1 open spot
        self.nodemgr.max_clients = 1
        # ensure we have an address that can be connected to
        node.NeoNode.addresses = [address.NetworkAddress(address="127.0.0.1:1111")]

        with self.assertLogs(network_logger, "DEBUG") as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 1 open pool spots", context.output[0])
        self.assertIn("Adding 127.0.0.1:1111 to connection queue", context.output[1])

    async def test_fill_open_connection_spots_dont_queue_addr_that_is_already_queued(
        self,
    ):
        self.node1 = node.NeoNode(object(), object())

        # ensure we have an address that can be connected to
        addr = address.NetworkAddress(address="127.0.0.1:1111")
        # create 1 open spot
        self.nodemgr.max_clients = 2
        node.NeoNode.addresses = [addr]
        # and ensure we pretend we have already queued this address
        self.nodemgr.queued_addresses = [addr]

        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch("asyncio.create_task"):
                await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 1 open pool spots", context.output[0])
        # context.output length should only contain the above debug message, nothing more
        self.assertEqual(1, len(context.output))

    async def test_fill_open_connection_spots_no_addresses_to_fill_spots(self):
        # we first test that no errors occur if we have open spots, but we still meet our minimum required clients
        # this should not cause any error count increases
        self.nodemgr.min_clients = 1
        self.nodemgr.max_clients = 2
        # just a placeholder to have a count matching min_clients
        self.nodemgr.nodes = [object()]
        node.NeoNode.addresses = []

        with self.assertLogs(network_logger, "DEBUG") as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn(
            "Found 1 open pool spots, trying to add nodes...", context.output[0]
        )
        self.assertIn(
            "No addresses available to fill spots. However, minimum clients still satisfied",
            context.output[1],
        )

        # next we clear the nodes and start inducing errors for not being able to fill the open spots
        self.nodemgr.nodes = []
        with self.assertLogs(network_logger, "DEBUG") as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn(
            "Found 2 open pool spots, trying to add nodes...", context.output[0]
        )
        self.assertIn("Increasing pool spot error count to 1", context.output[1])

        with self.assertLogs(network_logger, "DEBUG") as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn(
            "Found 2 open pool spots, trying to add nodes...", context.output[0]
        )
        self.assertIn("Increasing pool spot error count to 2", context.output[1])

        # 3rd time we reached our threshold
        # let's also set up an address in POOR state that should be reset to NEW
        addr = address.NetworkAddress(address="127.0.0.1:1111")
        addr.set_state_poor()
        node.NeoNode.addresses = [addr]
        with self.assertLogs(network_logger, "DEBUG") as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn(
            "Found 2 open pool spots, trying to add nodes...", context.output[0]
        )
        self.assertIn("Recycling old addresses", context.output[1])
        self.assertEqual(0, self.nodemgr.MAX_NODE_POOL_ERROR_COUNT)
        self.assertTrue(node.NeoNode.addresses[0].is_state_new)

    async def test_fill_open_connection_spots_node_timeout_error(self):
        self.nodemgr.min_clients = 1
        self.nodemgr.max_clients = 2

        addr = address.NetworkAddress(address="127.0.0.1:1111")
        node.NeoNode.addresses = [addr]

        call_back_result = None

        def client_connect_done(node_instance, failure):
            nonlocal call_back_result
            call_back_result = failure

        # listen to a connection done event
        msgrouter.on_client_connect_done += client_connect_done

        with self.assertLogs(network_logger, "DEBUG") as context:
            # for some reason patching asyncio.open_connection just doesn't work in this test
            # working around it like this
            with mock.patch.object(
                node.NeoNode, "connect_to", return_value=(None, (f"...", "Timed out"))
            ):
                await self.nodemgr._fill_open_connection_spots()

        # advance loop
        await asyncio.sleep(0.1)
        self.assertEqual("Timed out", call_back_result[1])
        self.assertIn(
            "Found 2 open pool spots, trying to add nodes...", context.output[0]
        )


class NodeManagerTimedTestCase(IsolatedAsyncioTestCase):  # asynctest.ClockedTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.nodemgr = nodemanager.NodeManager()

    def setUp(self) -> None:
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()
        settings.reset_settings_to_default()

        fake_reader, fake_writer = object(), object()
        self.node1 = node.NeoNode(fake_reader, fake_writer)
        self.node2 = node.NeoNode(fake_reader, fake_writer)

        self.addr1 = address.NetworkAddress(address="127.0.0.1:10333")
        self.addr2 = address.NetworkAddress(address="127.0.0.2:20333")
        self.node1.address = self.addr1
        self.node2.address = self.addr2

    @skip("Only worked with asynctest.ClockedTestCase, no replacement yet")
    async def test_utility_run_in_loop(self):
        counter = 0

        async def coro():
            nonlocal counter
            counter += 1

        self.nodemgr._run_in_loop(coro, 1)
        # trigger the task by ticking the loop once
        await self.advance(0)
        self.assertEqual(1, counter)
        await self.advance(0.9)
        self.assertEqual(1, counter)
        await self.advance(0.1)
        self.assertEqual(2, counter)
        self.nodemgr.tasks[0].cancel()

    async def test_monitor_node_height_disconnect_when_treshold_exceeded(self):
        # ensure we exceed the threshold
        self.node1.best_height_last_update = (
            datetime.utcnow().timestamp() - self.nodemgr.MAX_HEIGHT_UPDATE_DURATION - 1
        )
        self.nodemgr.nodes = [self.node1]
        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch.object(self.node1, "disconnect") as patched_node_disconnect:
                await self.nodemgr._monitor_node_height()
        self.assertIn("max height update threshold exceeded", context.output[0])
        patched_node_disconnect.assert_called_with(
            reason=address.DisconnectReason.POOR_PERFORMANCE
        )

    @skip("too slow, need pytest + looptime to enable")
    async def test_monitor_node_height_within_limits(self):
        # Note https://github.com/nolar/looptime/tree/main/looptime could solve this
        # if we were using pytest instead of unittest

        # test that if the last height update timestamp of a node is within the treshold
        # that we only ask for another update
        self.nodemgr.nodes = [self.node1]
        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch.object(
                self.node1, "send_message", side_effect=mock.AsyncMock()
            ) as patched_node_send_msg:
                await self.nodemgr._monitor_node_height()
        self.assertIn(" to send us a height update (PING)", context.output[0])
        self.assertGreater(len(patched_node_send_msg.call_args), 0)
        msg = patched_node_send_msg.call_args[0][0]
        self.assertEqual(message.MessageType.PING, msg.type)
        # finally test that our tasks is created
        self.assertEqual(1, len(self.nodemgr.tasks))
        await asyncio.sleep(1)
        # and the call back called once done.
        self.assertEqual(0, len(self.nodemgr.tasks))

    @skip("too slow, need pytest + looptime to enable")
    async def test_query_addresses(self):
        self.nodemgr.nodes = [self.node1]
        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch.object(
                self.node1, "request_address_list", side_effect=mock.AsyncMock()
            ) as patched_node_req_addr:
                await self.nodemgr._query_addresses()
        self.assertIn("Asking node ", context.output[1])
        self.assertIn("for its address list", context.output[1])
        patched_node_req_addr.assert_called_once()
        # finally test that our tasks is created
        self.assertEqual(1, len(self.nodemgr.tasks))
        await asyncio.sleep(1)
        # and the call back called once done.
        self.assertEqual(0, len(self.nodemgr.tasks))

    async def test_seed_list_processing(self):
        # specific testing of named addresses
        settings.network.seedlist = ["google.com:1111"]

        # for whatever reason I can't find the right way to mock the property, tried PropertyMock and what not.
        # this works *shrug*
        class wtf:
            host = "127.0.0.123"

        result = [wtf()]
        with mock.patch(
            "aiodns.DNSResolver.query",
            side_effect=mock.AsyncMock(return_value=result),
        ) as query_result:
            await self.nodemgr._process_seed_list_addresses()
        self.assertEqual(1, len(node.NeoNode.addresses))

        with self.assertLogs(network_logger, "DEBUG") as context:
            with mock.patch(
                "aiodns.DNSResolver.query", side_effect=aiodns.error.DNSError
            ) as query_result:
                await self.nodemgr._process_seed_list_addresses()
        self.assertIn(
            "Skipping google.com, address could not be resolved", context.output[0]
        )



import unittest
import json
from copy import deepcopy
from neo3.contracts import abi, manifest, nef, contract
from neo3.core import cryptography, types, serialization


class ContractGroupTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        private_key = b"\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        cls.keypair = cryptography.KeyPair(private_key)

        # capture from C#
        cls.expected_json = {
            "pubkey": "033d523f36a732974c0f7dbdfafb5206ecd087211366a274190f05b86d357f4bad",
            "signature": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==",
        }

    def shortDescription(self) -> str | None:
        # disable comment printing
        return None

    def test_is_valid(self):
        """
        var private_key = new byte[]
        {
            0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1
        };
        var bad_signature = new byte[]
        {
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0
        };
        var kp = new KeyPair(private_key);
        var cg = new ContractGroup() { PubKey = kp.PublicKey, Signature = bad_signature};
        Console.Write(cg.ToJson());
        """
        bad_signature = b"\x00" * 64
        cg = manifest.ContractGroup(self.keypair.public_key, bad_signature)
        self.assertFalse(cg.is_valid(types.UInt160.zero()))

        # finally test is_valid() with a keypair we know the private key off
        contract_hash = b"\x01" * 20
        signature = cryptography.sign(contract_hash, self.keypair.private_key)

        cg2 = manifest.ContractGroup(self.keypair.public_key, signature)
        self.assertTrue(cg2.is_valid(types.UInt160(contract_hash)))

    def test_json(self):
        bad_signature = b"\x00" * 64
        cg = manifest.ContractGroup(self.keypair.public_key, bad_signature)
        self.assertEqual(self.expected_json, cg.to_json())

        # if the test_to_json() passes then we know our manually created ControlGroup (cg) is valid,
        # we can now use that to validate our `from_json` object
        cg2 = manifest.ContractGroup.from_json(self.expected_json)
        self.assertEqual(cg.public_key, cg2.public_key)
        self.assertEqual(cg.signature, cg2.signature)

    def test_eq(self):
        cg = manifest.ContractGroup.from_json(self.expected_json)
        cg2 = manifest.ContractGroup.from_json(self.expected_json)
        self.assertFalse(cg == object())
        self.assertTrue(cg == cg2)


class ContractPermissionTestCase(unittest.TestCase):
    def test_default(self):
        cp = manifest.ContractPermission.default_permissions()
        self.assertTrue(cp.contract.is_wildcard)
        self.assertTrue(cp.methods.is_wildcard)

    def test_is_allowed_based_on_hash(self):
        # We test the group permissions where all methods are allowed to be called
        # if contract_hash is valid.
        dummy_contract_hash = types.UInt160.zero()
        contract_state = contract.ContractState(
            1, nef.NEF(), manifest.ContractManifest(), 0, dummy_contract_hash
        )

        # setup an allowed permission for a contract with UInt160.zero hash for all methods
        cpd = manifest.ContractPermissionDescriptor(contract_hash=dummy_contract_hash)
        cp = manifest.ContractPermission(
            contract=cpd, methods=manifest.WildcardContainer.create_wildcard()
        )
        self.assertTrue(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "dummy_method")
        )

        # now create a different contract hash and verify it does not give permission
        contract_state.hash = types.UInt160(b"\x01" * 20)
        self.assertFalse(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "dummy_method")
        )

    def test_is_allowed_based_on_group(self):
        # We test the group permissions where all methods are allowed to be called
        # if the 'groups' member is valid.

        private_key = b"\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        keypair = cryptography.KeyPair(private_key)

        dummy_contract_hash = types.UInt160.from_string("01" * 20)
        contract_state = contract.ContractState(
            1, nef.NEF(), manifest.ContractManifest(), 0, dummy_contract_hash
        )

        signature = cryptography.sign(
            contract_state.hash.to_array(), keypair.private_key
        )
        contract_state.manifest.groups = [
            manifest.ContractGroup(keypair.public_key, signature)
        ]

        cpd = manifest.ContractPermissionDescriptor(group=keypair.public_key)
        cp = manifest.ContractPermission(
            contract=cpd, methods=manifest.WildcardContainer.create_wildcard()
        )
        self.assertTrue(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "dummy_method")
        )

        # now modify the manifest to have a different `groups` attribute such that validation fails
        public_key = cryptography.ECPoint.deserialize_from_bytes(
            b"\x00"
        )  # ECPoint.Infinity
        contract_state.manifest.groups = [manifest.ContractGroup(public_key, signature)]
        self.assertFalse(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "dummy_method")
        )

    def test_is_allowed_invalid_method(self):
        # in the above tests we validated the 'group' and 'contract_hash' matching logic
        # now we validate 'method' matching
        dummy_contract_hash = types.UInt160.from_string("01" * 20)
        contract_state = contract.ContractState(
            1, nef.NEF(), manifest.ContractManifest(), 0, dummy_contract_hash
        )

        # setup an allowed permission for a contract with UInt160.zero hash for 2 methods
        cpd = manifest.ContractPermissionDescriptor(contract_hash=dummy_contract_hash)
        cp = manifest.ContractPermission(
            contract=cpd,
            methods=manifest.WildcardContainer(data=["method1", "method2"]),
        )
        self.assertTrue(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "method1")
        )
        self.assertTrue(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "method2")
        )
        self.assertFalse(
            cp.is_allowed(contract_state.hash, contract_state.manifest, "method3")
        )

    def test_to_json(self):
        # var cg = ContractPermission.DefaultPermission;
        # Console.WriteLine(cg.ToJson());
        cp = manifest.ContractPermission.default_permissions()
        expected = {"contract": "*", "methods": "*"}
        self.assertDictEqual(expected, cp.to_json())

    def test_from_json(self):
        # we use data from 'test_to_json'
        # we only test parsing basic data, as the detailed parsing is covered
        # by the tests for specific object classes
        json = {"contract": "*", "methods": "*"}
        from_json = manifest.ContractPermission.from_json(json)
        self.assertTrue(from_json.methods.is_wildcard)
        self.assertTrue(from_json.contract.is_wildcard)

    def test_eq(self):
        json = {"contract": "*", "methods": "*"}
        from_json = manifest.ContractPermission.from_json(json)
        from_json2 = manifest.ContractPermission.from_json(json)
        self.assertFalse(from_json == object())
        self.assertTrue(from_json == from_json2)


class WildcardContainerTestCase(unittest.TestCase):
    def test_dunders(self):
        wc = manifest.WildcardContainer(data=["method1", "method2"])
        wc2 = manifest.WildcardContainer(data=["method1", "method2"])
        self.assertIn("method1", wc)
        self.assertIn("method2", wc)
        self.assertNotIn("method3", wc)
        self.assertEqual("method2", wc[1])
        self.assertEqual(2, len(wc))
        self.assertNotEqual(wc, object())
        self.assertEqual(wc, wc2)

    def test_wildcard(self):
        wc = manifest.WildcardContainer.create_wildcard()
        self.assertTrue(wc.is_wildcard)

        wc = manifest.WildcardContainer(["method1"])
        self.assertFalse(wc.is_wildcard)

    def test_to_json(self):
        wc = manifest.WildcardContainer.create_wildcard()
        self.assertDictEqual({"wildcard": "*"}, wc.to_json())

        wc = manifest.WildcardContainer(data=["method1", "method2"])
        self.assertDictEqual({"wildcard": ["method1", "method2"]}, wc.to_json())

    def test_from_json_default(self):
        wc = manifest.WildcardContainer.from_json({"wildcard": "*"})
        self.assertTrue(wc.is_wildcard)

        wc = manifest.WildcardContainer.from_json({"wildcard": ["method1", "method2"]})
        self.assertFalse(wc.is_wildcard)
        self.assertIn("method1", wc)
        self.assertIn("method2", wc)

        with self.assertRaises(ValueError) as context:
            manifest.WildcardContainer.from_json({})
        self.assertEqual(
            "Invalid JSON - Cannot recreate wildcard from None", str(context.exception)
        )

        with self.assertRaises(ValueError) as context:
            manifest.WildcardContainer.from_json({"wildcard": "abc"})
        self.assertEqual(
            "Invalid JSON - Cannot deduce WildcardContainer type from: abc",
            str(context.exception),
        )

    def test_from_json_as_type(self):
        wc = manifest.WildcardContainer.from_json_as_type(
            {"wildcard": "*"}, lambda: None
        )
        self.assertTrue(wc.is_wildcard)

        t1 = types.UInt160.zero()
        t2 = types.UInt160.from_string("11" * 20)
        t3 = types.UInt160.from_string("22" * 20)
        wc = manifest.WildcardContainer.from_json_as_type(
            {"wildcard": [str(t1), str(t2)]}, lambda t: types.UInt160.from_string(t)
        )
        self.assertFalse(wc.is_wildcard)
        self.assertIn(t1, wc)
        self.assertIn(t2, wc)
        self.assertNotIn(t3, wc)

        with self.assertRaises(ValueError) as context:
            manifest.WildcardContainer.from_json_as_type({}, lambda: None)
        self.assertEqual(
            "Invalid JSON - Cannot recreate wildcard from None", str(context.exception)
        )

        with self.assertRaises(ValueError) as context:
            manifest.WildcardContainer.from_json_as_type(
                {"wildcard": "abc"}, lambda: None
            )
        self.assertEqual(
            "Invalid JSON - Cannot deduce WildcardContainer type from: abc",
            str(context.exception),
        )


class ManifestTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var manifest = new ContractManifest()
        {
            Name = "test_contract",
            Groups = new ContractGroup[0],
            SupportedStandards = Array.Empty<string>(),
            Abi = new ContractAbi()
            {
                Events = new ContractEventDescriptor[0],
                Methods = new ContractMethodDescriptor[]
                {
                    new ContractMethodDescriptor {
                        Name = "main_entry",
                        Offset = 0,
                        Parameters = new ContractParameterDefinition[0],
                        ReturnType = ContractParameterType.Integer,
                        Safe = true
                    }
                }
            },
            Permissions = new[] { ContractPermission.DefaultPermission },
            Trusts = WildcardContainer<ContractPermissionDescriptor>.Create(),
            Extra = null
        };
        Console.WriteLine($"{manifest.ToJson()}");
        """
        cls.expected_json = {
            "name": "test_contract",
            "groups": [],
            "features": {},
            "supportedstandards": [],
            "abi": {
                "methods": [
                    {
                        "name": "main_entry",
                        "parameters": [],
                        "returntype": "Integer",
                        "offset": 0,
                        "safe": True,
                    }
                ],
                "events": [],
            },
            "permissions": [{"contract": "*", "methods": "*"}],
            "trusts": [],
            "extra": None,
        }

    def test_create(self):
        cm = manifest.ContractManifest("test_contract")
        method1 = abi.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=abi.ContractParameterType.INTEGER,
            safe=True,
        )
        cm.abi.methods = [method1]
        self.assertEqual(self.expected_json, cm.to_json())

    def test_serialize(self):
        # if test_create() passes, then we know `to_json()` is ok, which serialize internally uses
        cm = manifest.ContractManifest("test_contract")
        method1 = abi.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=abi.ContractParameterType.INTEGER,
            safe=True,
        )
        cm.abi.methods = [method1]
        with serialization.BinaryReader(cm.to_array()) as br:
            data = br.read_var_string()
        self.assertDictEqual(self.expected_json, json.loads(data))

    def test_to_json_with_trusts_extra(self):
        # create a default manifest
        m = manifest.ContractManifest("test_contract")
        method1 = abi.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=abi.ContractParameterType.INTEGER,
            safe=True,
        )
        m.abi.methods = [method1]

        t1 = manifest.ContractPermissionDescriptor(
            contract_hash=types.UInt160.from_string("01" * 20)
        )
        t2 = manifest.ContractPermissionDescriptor(
            contract_hash=types.UInt160.from_string("02" * 20)
        )
        m.trusts = manifest.WildcardContainer(data=[t1, t2])
        m.extra = False
        json_out = m.to_json()
        self.assertIn(t1.to_json()["contract"], json_out["trusts"])
        self.assertIn(t2.to_json()["contract"], json_out["trusts"])
        self.assertFalse(json_out["extra"])

    def test_from_json(self):
        expected_json = deepcopy(self.expected_json)
        cm = manifest.ContractManifest.from_json(expected_json)
        default = manifest.ContractManifest()
        self.assertEqual(default.groups, cm.groups)
        self.assertEqual(default.permissions, cm.permissions)
        self.assertEqual(default.trusts, cm.trusts)
        self.assertEqual(default.extra, cm.extra)

    def test_deserialization(self):
        # this assumes test_serialization() passes
        cm = manifest.ContractManifest("test_contract")
        method1 = abi.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=abi.ContractParameterType.INTEGER,
            safe=True,
        )
        cm.abi.methods = [method1]
        cm_deserialized = manifest.ContractManifest.deserialize_from_bytes(
            cm.to_array()
        )
        self.assertEqual(cm, cm_deserialized)

    def test_is_valid(self):
        # create a contract
        cm = manifest.ContractManifest("test_contract")
        method1 = abi.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=abi.ContractParameterType.INTEGER,
            safe=True,
        )
        cm.abi.methods = [method1]

        # A contract hash is normally created from a combination of transaction sender (UInt160), NEF checksum and manifest.name
        # For this test we'll keep it simple and pretend that hash is all zeros
        dummy_contract_hash = types.UInt160.zero()

        # a "default" contract has no groups data, which is the same as always allow
        self.assertTrue(cm.is_valid(dummy_contract_hash))

        # now try to add a malicious group member (meaning; the member did not actually sign the ABI)
        private_key = b"\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        keypair = cryptography.KeyPair(private_key)
        bad_signature = bytes(64)
        cm.groups = [manifest.ContractGroup(keypair.public_key, bad_signature)]
        # this time validation should fail
        self.assertFalse(cm.is_valid(dummy_contract_hash))

        # Finally test with a group member that did sign the ABI
        good_signature = cryptography.sign(
            dummy_contract_hash.to_array(), keypair.private_key
        )
        cm.groups = [manifest.ContractGroup(keypair.public_key, good_signature)]
        self.assertTrue(cm.is_valid(dummy_contract_hash))

    def test_eq(self):
        cm = manifest.ContractManifest.from_json(self.expected_json)
        cm2 = manifest.ContractManifest.from_json(self.expected_json)
        self.assertFalse(cm == object())
        self.assertTrue(cm == cm2)


class ContractPermissionDescriptorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        private_key = b"\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        cls.keypair = cryptography.KeyPair(private_key)

    def test_wildcard(self):
        cpd = manifest.ContractPermissionDescriptor()
        self.assertTrue(cpd.is_wildcard)
        self.assertDictEqual({"contract": "*"}, cpd.to_json())
        cpd_from_json = manifest.ContractPermissionDescriptor.from_json(cpd.to_json())
        self.assertEqual(cpd.contract_hash, cpd_from_json.contract_hash)
        self.assertEqual(cpd.group, cpd_from_json.group)

    def test_group(self):
        cpd = manifest.ContractPermissionDescriptor(contract_hash=types.UInt160.zero())
        self.assertTrue(cpd.is_hash)
        self.assertFalse(cpd.is_group)
        self.assertFalse(cpd.is_wildcard)
        self.assertDictEqual(
            {"contract": "0x0000000000000000000000000000000000000000"}, cpd.to_json()
        )
        cpd_from_json = manifest.ContractPermissionDescriptor.from_json(cpd.to_json())
        self.assertEqual(cpd.contract_hash, cpd_from_json.contract_hash)
        self.assertEqual(cpd.group, cpd_from_json.group)

    def test_contract_hash(self):
        cpd = manifest.ContractPermissionDescriptor(group=self.keypair.public_key)
        self.assertFalse(cpd.is_hash)
        self.assertTrue(cpd.is_group)
        self.assertFalse(cpd.is_wildcard)
        self.assertDictEqual(
            {
                "contract": "033d523f36a732974c0f7dbdfafb5206ecd087211366a274190f05b86d357f4bad"
            },
            cpd.to_json(),
        )
        cpd_from_json = manifest.ContractPermissionDescriptor.from_json(cpd.to_json())

        self.assertEqual(cpd.contract_hash, cpd_from_json.contract_hash)
        self.assertEqual(cpd.group, cpd_from_json.group)

    def test_exceptions(self):
        # test construction with too many arguments given
        with self.assertRaises(ValueError) as context:
            cpd = manifest.ContractPermissionDescriptor(
                contract_hash=types.UInt160.zero(), group=self.keypair.public_key
            )
        self.assertIn("Maximum 1 argument", str(context.exception))

        # test from_json with invalid json
        with self.assertRaises(ValueError) as context:
            manifest.ContractPermissionDescriptor.from_json({"contract": None})
        self.assertEqual(
            "Invalid JSON - Cannot deduce permission type from None",
            str(context.exception),
        )

        with self.assertRaises(ValueError) as context:
            manifest.ContractPermissionDescriptor.from_json({"contract": "abc"})
        self.assertEqual(
            "Invalid JSON - Cannot deduce permission type from: abc",
            str(context.exception),
        )



import unittest
from neo3.network.ipfilter import IPFilter


class IPFilteringTestCase(unittest.TestCase):
    def test_nobody_allowed(self):
        filter = IPFilter()
        filter.load_config({"blacklist": ["0.0.0.0/0"], "whitelist": []})

        self.assertFalse(filter.is_allowed("127.0.0.1"))
        self.assertFalse(filter.is_allowed("10.10.10.10"))

    def test_nobody_allowed_except_one(self):
        filter = IPFilter()
        filter.load_config({"blacklist": ["0.0.0.0/0"], "whitelist": ["10.10.10.10"]})

        self.assertFalse(filter.is_allowed("127.0.0.1"))
        self.assertFalse(filter.is_allowed("10.10.10.11"))
        self.assertTrue(filter.is_allowed("10.10.10.10"))

    def test_everybody_allowed(self):
        filter = IPFilter()
        filter.load_config({"blacklist": [], "whitelist": []})

        self.assertTrue(filter.is_allowed("127.0.0.1"))
        self.assertTrue(filter.is_allowed("10.10.10.11"))
        self.assertTrue(filter.is_allowed("10.10.10.10"))

        filter.load_config({"blacklist": [], "whitelist": ["0.0.0.0/0"]})

        self.assertTrue(filter.is_allowed("127.0.0.1"))
        self.assertTrue(filter.is_allowed("10.10.10.11"))
        self.assertTrue(filter.is_allowed("10.10.10.10"))

        filter.load_config({"blacklist": ["0.0.0.0/0"], "whitelist": ["0.0.0.0/0"]})

        self.assertTrue(filter.is_allowed("127.0.0.1"))
        self.assertTrue(filter.is_allowed("10.10.10.11"))
        self.assertTrue(filter.is_allowed("10.10.10.10"))

    def test_everybody_allowed_except_one(self):
        filter = IPFilter()
        filter.load_config({"blacklist": ["127.0.0.1"], "whitelist": []})

        self.assertFalse(filter.is_allowed("127.0.0.1"))
        self.assertTrue(filter.is_allowed("10.10.10.11"))
        self.assertTrue(filter.is_allowed("10.10.10.10"))

    def test_disallow_ip_range(self):
        filter = IPFilter()
        filter.load_config({"blacklist": ["127.0.0.0/24"], "whitelist": []})

        self.assertFalse(filter.is_allowed("127.0.0.0"))
        self.assertFalse(filter.is_allowed("127.0.0.1"))
        self.assertFalse(filter.is_allowed("127.0.0.100"))
        self.assertFalse(filter.is_allowed("127.0.0.255"))
        self.assertTrue(filter.is_allowed("10.10.10.11"))
        self.assertTrue(filter.is_allowed("10.10.10.10"))

    def test_updating_blacklist(self):
        filter = IPFilter()
        filter.load_config({"blacklist": [], "whitelist": []})

        self.assertTrue(filter.is_allowed("127.0.0.1"))

        filter.blacklist_add("127.0.0.0/24")
        self.assertFalse(filter.is_allowed("127.0.0.1"))
        # should have no effect, only exact matches
        filter.blacklist_remove("127.0.0.1")
        self.assertFalse(filter.is_allowed("127.0.0.1"))

        filter.blacklist_remove("127.0.0.0/24")
        self.assertTrue(filter.is_allowed("127.0.0.1"))

    def test_updating_whitelist(self):
        filter = IPFilter()
        filter.load_config({"blacklist": ["0.0.0.0/0"], "whitelist": []})

        self.assertFalse(filter.is_allowed("127.0.0.1"))

        filter.whitelist_add("127.0.0.0/24")
        self.assertTrue(filter.is_allowed("127.0.0.1"))

        filter.whitelist_remove("127.0.0.1")
        # should have no effect, only exact matches
        self.assertTrue(filter.is_allowed("127.0.0.1"))

        filter.whitelist_remove("127.0.0.0/24")
        self.assertFalse(filter.is_allowed("127.0.0.1"))

    def test_invalid_config_loading(self):
        filter = IPFilter()
        # mandatory keys not present
        with self.assertRaises(ValueError) as ctx:
            filter.load_config({"blacklist": []})
        self.assertEqual("whitelist key not found", str(ctx.exception))

        # mandatory key blacklist not present
        with self.assertRaises(ValueError) as ctx:
            filter.load_config({"whitelist": []})
        self.assertEqual("blacklist key not found", str(ctx.exception))

    def test_config_reset(self):
        filter = IPFilter()
        filter.blacklist_add("127.0.0.1")
        self.assertEqual(1, len(filter._config["blacklist"]))
        filter.reset()
        self.assertEqual(filter.default_config, filter._config)
